from ._segment_functions import (
    closest_point_of_line_segment,
    subdivide_segment,
    subdivide_segments,
)

__all__ = ["closest_point_of_line_segment", "subdivide_segment", "subdivide_segments"]
